from collections import deque

import numpy as np
import pandas as pd
import scipy.sparse
import scipy.spatial
import scipy.stats
import sklearn
from tqdm import tqdm

from nest import get_neighbor_adjacency


class GaussianExpressionModel:
    def __init__(self, dim=16, pca=True):
        self.dim = dim
        self.use_pca = pca

        # initialized later
        self.means = None
        self.covs = None

        self.regions = None
        self.adata = None

    def set_data(self, adata, regions):
        self.regions = regions
        self.adata = adata

        # perform PCA on genes
        try:
            X = np.array(self.adata.X.toarray())
        except AttributeError:
            X = self.adata.X
        if self.use_pca:
            self.pca_transform = sklearn.decomposition.PCA(n_components=self.dim)
            self.pca_transform.fit(X)
            self.pca = self.pca_transform.transform(X)
        else:
            if adata.shape[1] != self.dim:
                raise ValueError
            self.pca = X

        self.pca_t = np.transpose(self.pca)

    def update_data(self, adata):
        self.adata = adata
        # apply the PCA transformation, but without re-fitting
        X = self.adata.X.log1p().toarray()
        self.pca = self.pca_transform.transform(X)
        self.pca_t = np.transpose(self.pca)

    def update_parameters(self, gamma):
        gamma_sum = np.sum(gamma, axis=0)

        gamma_t = np.transpose(gamma)
        self.means = (gamma_t @ self.pca) / gamma_sum.reshape(self.regions, 1)

        pca_minus_mean = self.pca.reshape((-1, 1, self.dim)) - self.means.reshape(
            (1, self.regions, self.dim))
        A = pca_minus_mean.reshape(-1, self.regions, self.dim, 1) * \
            pca_minus_mean.reshape(-1, self.regions, 1, self.dim)

        self.covs = np.einsum('ij,jimn->imn', gamma_t, A) / gamma_sum.reshape((self.regions, 1, 1))

    def compute_log_likelihood(self, gamma):
        # does not use gamma
        ll = np.zeros(shape=gamma.shape)
        for i in range(self.regions):
            ll[:, i] = scipy.stats.multivariate_normal.logpdf(x=self.pca, mean=self.means[i, :],
                                                              cov=self.covs[i, :, :],
                                                              allow_singular=True)
        return ll


class HMRF:
    """
    Likelihood model for HMRF
    """

    def __init__(self, beta, threshold=None, k=None, cap=None):
        self.threshold = threshold
        self.k = k
        self.beta = beta
        self.cap = cap

        # To be set later
        self.adjacency = None
        self.regions = None

    def set_data(self, adata, regions):
        self.regions = regions
        coords = adata.obsm['spatial']

        # This function automatically switches depending on which of threshold/k is not None
        self.adjacency = get_neighbor_adjacency(coords, eps=self.threshold, k=self.k)

    def compute_log_likelihood(self, gamma):
        log_likelihood = (self.adjacency @ gamma)
        if self.cap is not None:
            log_likelihood[log_likelihood > self.cap] = self.cap

        log_likelihood = self.beta * log_likelihood
        return log_likelihood

    def update_parameters(self, gamma):
        pass


class HMRFSegmentationModel:
    def __init__(self, adata, regions, eps=None, k=None, model_weights=None, label_name=None,
                 max_iterations=200, verbose=False, dim=16, pca=True, beta=1):
        self.adata = adata
        self.likelihood_models = [GaussianExpressionModel(dim=dim, pca=pca),
                                  HMRF(beta=beta, threshold=eps, k=k)]
        # number of regions to segment into
        self.regions = regions
        self.model_weights = model_weights
        self.label_name = label_name
        self.max_iterations = max_iterations
        self.verbose = verbose

        # Set in fit function

        self.gamma = None
        self.converged = False

        self.model_ll = None
        self.class_ll = None

        for model in self.likelihood_models:
            model.set_data(adata, self.regions)

    def fit(self, max_iterations=None, update_labels=False, tol=1e-6, metric=None, verbose=False,
            **kwargs):
        ncells = len(self.adata)

        # if no current class information, run initialization
        if self.gamma is None:
            self.initialize_gamma()

        last_10_ll = deque()

        flag = False  # convergence stopping

        if max_iterations is not None:
            iterations = max_iterations
        else:
            iterations = self.max_iterations

        if verbose:
            pbar = tqdm(range(iterations), leave=False)
        else:
            pbar = range(iterations)

        for it in pbar:
            ll = np.zeros(shape=(ncells, self.regions))

            # Compute log-likelihood summed across all models
            for i, model in enumerate(self.likelihood_models):
                model.update_parameters(self.gamma)

                if self.model_weights is not None:
                    weight = self.model_weights[i]
                else:
                    weight = 1

                ll += weight * model.compute_log_likelihood(self.gamma)

            # Also do class size-based likelihood
            class_prob = np.mean(self.gamma, axis=0)
            ll += np.log(class_prob).reshape(1, self.regions)

            # Compute new class probabilities
            self.ll = ll
            self.gamma = scipy.special.softmax(ll, axis=1)

            self.model_ll = np.mean(ll * self.gamma)

            if tol is not None and it >= 10 and self.model_ll - max(last_10_ll) < tol:
                flag = True

            if len(last_10_ll) >= 10:
                last_10_ll.popleft()
            last_10_ll.append(self.model_ll)

            # compute overall model likelihood to watch for convergence
            self.class_ll = ll[:, self.class_labels].mean()

            if self.verbose:
                pbar.set_description(f"LL: {self.model_ll}")

            if flag:
                self.converged = True
                if self.verbose:
                    pbar.set_description("Converged: LL %f" % self.model_ll)
                break

        if update_labels:
            self.adata.obs[self.label_name] = pd.Categorical(self.class_labels)

    def initialize_gamma(self):
        try:
            data_raw = self.adata.X.toarray()
        except (TypeError, AttributeError) as e:
            data_raw = self.adata.X

        self.kmeans_initialization(data_raw)

    def kmeans_initialization(self, data):
        # do dimensionality reduction with PCA
        reduced_dims = 16
        pca = sklearn.decomposition.PCA(n_components=reduced_dims)
        data = pca.fit_transform(data)

        class_labels = sklearn.cluster.KMeans(n_clusters=self.regions).fit(data).labels_
        labels_onehot = pd.get_dummies(class_labels, columns=range(self.regions)).to_numpy()

        # how much uncertainty to introduce in initial clusters
        eps = 0.1

        self.gamma = (1 - eps) * labels_onehot + eps / self.regions * np.ones(
            shape=labels_onehot.shape)

    @property
    def class_labels(self):
        labels = np.argmax(self.gamma, axis=1)
        return labels



import os

import matplotlib.pyplot as plt
from matplotlib import colors
import scanpy as sc
import seaborn as sns
import numpy as np
import pandas as pd
import colorcet as cc

# Global variables store dataset-specific information
from nest.hotspot.geometry import similarity_map

library_id = None
img = None
scale = None
spot_size_preset = None
parameters_set = False
image_path = None


def set_dataset_plot_parameters(dataset):
    """
    Sets global variables that are used in spatial plots for non-Visium datasets
    :param dataset:
    :return:
    """
    global library_id, img, scale, spot_size_preset, parameters_set, image_path
    if dataset == "seqfish":
        img = None
        scale = None
        spot_size_preset = 0.045
    elif "merfish" in dataset:
        img = None
        scale = None
        spot_size_preset = 0.02
    elif "imc" in dataset:
        img = None
        scale = None
        spot_size_preset = 10
    elif "slideseq" in dataset:
        img = None
        scale = None
        spot_size_preset = 30
    elif "Spatial_LIBD" in dataset:
        # TODO: load the image
        img = None
        scale = None
        spot_size_preset = 100
    else:
        pass

    # So that we can check that the globals are set correctly later
    parameters_set = True


def spatial(adata, spot_size=None, **kwargs):
    """
    Wraps scanpy's pl.spatial function, using parameters from set_dataset_plot_parameters,
    in order to make it a bit
    less annoying to manage parameters when analyzing non-Visium data
    :param adata: adata object to plot (spatial information in `.obsm['spatial']`)
    :param spot_size: size of spots used to show color parameter
    :param kwargs: Other keyword arguments to pass to sc.pl.spatial, such as color=, etc.
    :return:
    """

    if spot_size is None:
        spot_size = spot_size_preset

    sc.pl.spatial(adata, img=img, scale_factor=scale, spot_size=spot_size,
                  **kwargs)

    try:
        ax = kwargs['ax']
        ax.set_xlabel('')
        ax.set_ylabel('')
    except (KeyError, AttributeError):
        pass


def volcano(df, interaction=None, gene_list=None, ax=None, show=True, xlim=None, ylim=None,
            axis_labels=True, fc_cutoff=0, p_cutoff=2, fontsize=None, xoffset=0, yoffset=0.5,
            marker=None, title=None,
            **kwargs):
    if interaction is not None:
        # filter out the actual genes from the interaction
        ligand, receptor = interaction.split("_")
        ligand = ligand.capitalize()
        receptor = receptor.capitalize()
        df = df.drop(index=ligand).drop(index=receptor)
    x = df["log2(fc)"]
    y = df["-log10(p)"]

    # for coloring points based on significance
    hue = np.zeros(shape=x.shape)
    hue[np.logical_and(x > fc_cutoff, y > p_cutoff)] = 1
    hue[np.logical_and(x < -fc_cutoff, y > p_cutoff)] = 2
    palette = []
    if 0 in hue:
        palette.append(sns.color_palette("muted")[7])
        # palette.append("gray")
    if 1 in hue:
        palette.append(sns.color_palette("muted")[3])
        # palette.append("red")
    if 2 in hue:
        palette.append(sns.color_palette("muted")[0])
        # palette.append("blue")

    # For indicating a subset of genes using a different marker
    style = None
    if marker is not None:
        style = np.zeros(len(x), dtype=np.int32)
        for idx in range(len(x)):
            style[idx] = df.index[idx] in marker

    ax = sns.scatterplot(x=x, y=y, hue=hue, palette=palette, legend=None, ax=ax, style=style,
                         markers=['o', '^'],
                         **kwargs)

    for i, row in df.iterrows():
        x = row["log2(fc)"]
        y = row["-log10(p)"]
        if gene_list is not None and (gene_list is True or i in gene_list):
            # gene_list can be either a list or a dictionary containing offsets
            try:
                gene_xoffset, gene_yoffset = gene_list[i]
            except TypeError:
                gene_xoffset, gene_yoffset = xoffset, yoffset
            ax.text(x + gene_xoffset, y + gene_yoffset, i, fontsize=fontsize,
                    horizontalalignment='right')

    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    if not axis_labels:
        ax.set_xlabel('')
        ax.set_ylabel('')
    else:
        ax.set_xlabel('log2(fc)', size=fontsize)
        ax.set_ylabel('-log10(p)', size=fontsize)
    plt.xticks(size=fontsize)
    plt.yticks(size=fontsize)

    # TODO: titles
    if title is not None:
        ax.set_title(title)
    elif interaction is not None:
        ax.set_title(" - ".join([v.capitalize() for v in interaction.split("_")]))
    if show:
        plt.show()

    return ax


def plot_similarity_map(adata, idx, ax=None, linewidth=0.5, linecolor="black", title="",
                        show=False, **kwargs):
    if ax is None:
        fig, ax = plt.subplots()
    try:
        scale_factor = list(adata.uns['spatial'].values())[0]['scalefactors'][
            'tissue_hires_scalef']
    except KeyError:
        if scale is not None:
            scale_factor = scale
        else:
            scale_factor = 1
    try:
        boundary = scale_factor * adata.uns["multi_boundaries"][str(idx)]
    except KeyError:
        raise ValueError("Need to compute boundaries first")
    #ax.fill(boundary[:, 0], boundary[:, 1], color_string)
    ax.plot(boundary[:, 0], boundary[:, 1], linewidth=linewidth, color=linecolor)

    arr = similarity_map(adata, idx=idx)

    adata.obs['tmp'] = arr
    spatial(adata, color="tmp", alpha_img=0.5, ax=ax, title=title, frameon=False, show=show,
            **kwargs)


# Construct hotspot tree
from matplotlib.patches import Rectangle, FancyBboxPatch, BoxStyle, Circle, Shadow
from matplotlib.lines import Line2D
from scipy.spatial.distance import jaccard
try:
    from functools import cache
except ImportError:
    # Python 3.8 compatibility
    def cache(func):
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper


def nested_structure_plot(adata, child_threshold=0.75, figsize=None, fontsize=None, legend=True, legend_ncol=None,
                          alpha_high=0.8, alpha_low=0.1, ax=None, legend_kwargs=None):
    if legend_kwargs is None:
        legend_kwargs = {}


    overlap, A = compute_ch_overlap_matrix(adata, child_threshold)
    row, col = np.where(A)
    # find top level of res
    top_level = np.array(list(set(range(len(A))) - set(np.where(A)[0])))
    # compute length based on normalized size
    top_level = {k: np.count_nonzero(pd.notnull(adata.obs[f'hotspots_multi_{k}'])) for k in top_level}
    total = np.sum(list(top_level.items()))
    top_level = {k: v / total for k, v in top_level.items()}
    #print(row, col)
    # reorder hotspots by depth to get the right arrangement
    @cache
    def get_depth(idx, start):
        cur_max = start
        for c in row[col == idx]:
            cur_max = max(cur_max, get_depth(c, start + 1))
        return cur_max

    def set_alpha(color, alpha):
        return colors.to_hex(colors.to_rgba(color, alpha=alpha), keep_alpha=True)

    num_elements = {int(k): np.count_nonzero(pd.notnull(adata.obs[f'hotspots_multi_{k}']))
                    for k, v in adata.uns['multi_hotspots'].items()}

    start_x = 0
    for k, v in sorted(top_level.items(), key=lambda x: -get_depth(x[0], 1)):
        color = adata.uns['multi_hotspots_colors'][str(k)]
        width = v
        top_level[k] = [start_x, width, set_alpha(color, alpha_high), True]
        if row[col == k].size > 0:
            addition = 0.01
        else:
            addition = 0
        start_x += width + addition

    # go to next level
    cur_level = top_level
    levels = [top_level]
    were_children = True

    depth_start = 2
    while were_children:
        next_level = {}
        were_children = False
        for k, (start_x, width, color, is_top) in cur_level.items():
            if not is_top:
                next_level[k] = [start_x, width, set_alpha(color, alpha_low), False]
                continue
            children = row[col == k]
            remaining_frac = 1

            start_x_sub = start_x
            for c in sorted(children, key=lambda x: -get_depth(x, depth_start)):
                were_children = True
                cur_child_frac = overlap[c, k] / num_elements[k]
                remaining_frac -= cur_child_frac
                color_sub = set_alpha(adata.uns['multi_hotspots_colors'][str(c)], 0.8)
                width_sub = cur_child_frac * width
                mul = 1.0
                next_level[c] = [start_x_sub, mul * width_sub, color_sub, True]
                start_x_sub += width_sub

            # draw the remaining bit (possibly the whole bit if no children) in reduced alpha
            if remaining_frac < 1:
                next_level[k] = [start_x_sub, width * remaining_frac, set_alpha(color, alpha_high), False]
            else:
                next_level[k] = [start_x_sub, width * remaining_frac, set_alpha(color, alpha_low), False]



        if not were_children:
            break

        levels.append(next_level)
        cur_level = next_level
        depth_start += 1

    # Bar plot of the coexpression hotspots
    start_y = 1
    height = 0.2

    if ax is None:
        _, ax = plt.subplots(figsize=figsize)
    ax.set_axis_off()

    for idx, level in enumerate(levels):
        for k, [start_x, width, color, is_top] in level.items():
            # patch = Rectangle([start_x, start_y], width, height, facecolor=color)
            rounding_size = min(0.01, width / 2)
            patch = FancyBboxPatch([start_x, start_y - height], width, height,
                                   boxstyle=BoxStyle("Round",
                                                     pad=-0.001,
                                                     rounding_size=rounding_size),
                                   facecolor=color, edgecolor=[1.0, 1.0, 1.0, 0.0])
            ax.add_patch(patch)
            if False and is_top and width > 0.02:
                ax.text(start_x + width / 2, start_y - 0.101, k, ha='center', va='center',
                        fontdict={'color': 'white', 'size': fontsize})
        ax.text(-0.01, start_y - height / 2, f'Layer {idx + 1}', fontdict={'size': fontsize}, ha='right', va='center')
        start_y -= 1.1 * height

    if legend:
        # TODO: make legend use the fontsize parameter
        legend_handles = []
        for k, color in adata.uns['multi_hotspots_colors'].items():
            legend_handles.append(Circle([0, 0], radius=5, color=color, label=str(k)))
        for k, v in  {'loc': "upper center", 'bbox_to_anchor': (1, 1),
                  'ncol': legend_ncol, 'frameon': False}.items():
            legend_kwargs.setdefault(k, v)

        ax.legend(handles=legend_handles, **legend_kwargs)


def compute_ch_overlap_matrix(adata, threshold, filter_indirect=True):
    num_ch = len(adata.uns['multi_hotspots'])

    # compute the element list for each CH
    element_list = {int(k): np.array(pd.notnull(adata.obs[f'hotspots_multi_{k}']))
                    for k, v in adata.uns['multi_hotspots'].items()}
    num_elements = {k: np.count_nonzero(v) for k, v in element_list.items()}
    overlap = np.zeros(shape=(num_ch, num_ch))
    A = np.zeros(shape=(num_ch, num_ch))
    for idx in range(num_ch):
        for idy in range(num_ch):
            if idx == idy:
                continue
            overlap[idx, idy] = np.count_nonzero(element_list[idx] & element_list[idy])
            A[idx, idy] = overlap[idx, idy] / num_elements[idx]

    A = A > threshold

    if filter_indirect:
        A = A & ~(A @ A)

    return overlap, A


import scanpy as sc


class CellChat:
    def __init__(self):
        # set up to run the R code
        try:
            from rpy2 import robjects
            from rpy2.robjects.packages import importr
            import anndata2ri
        except ImportError:
            raise ImportError("Optional dependencies rpy2 and anndata2ri must be installed to"
                              " run the Cellchat wrapper.")

        anndata2ri.activate()
        self.robjects = robjects
        self.anndata2ri = anndata2ri

        # TODO: figure out what error happens if cellchat is not installed correctly
        self.cellchat = importr("CellChat")

    def run(self, adata, group_by="class"):
        adata = adata.copy()
        # converter does not currently seem to support sparse matrix entries
        try:
            adata.X = adata.X.todense()
        except AttributeError:
            pass

        # adata.X = np.exp(adata.X) + 1

        # discard extraneous information (which may or may not convert into R)
        adata.obs = sc.get.obs_df(adata, keys=["class"])

        self.robjects.r('''
                    f <- function(s, group.by) {
                        assay(s, "logcounts") = assay(s, "X")
                        #print(s)
                        #print(dim(pos))
                        # Save this for testing analysis
                        save(s, file="/Users/blw/Documents/data/sce.dat")

                        cellchat <- createCellChat(object = s, group.by = group.by)

                        print(cellchat)

                        CellChatDB <- CellChatDB.mouse
                        library(dplyr)
                        dplyr::glimpse(CellChatDB$interaction)

                        CellChatDB.use <- subsetDB(CellChatDB, search = "Secreted Signaling")
                        cellchat@DB = CellChatDB.use

                        cellchat <- subsetData(cellchat)
                        cellchat <- identifyOverExpressedGenes(cellchat)
                        cellchat <- identifyOverExpressedInteractions(cellchat)
                        cellchat <- projectData(cellchat, PPI.mouse)

                        #prior.thresholds <- list("Secreted Signaling"=0.1)
                        #pos <- matrix(rnorm(13938*2), 13938)
                        # TODO: figure out the right way to get the stuff loaded
                        library(Matrix)
                        cellchat <- computeCommunProb(cellchat, #prior.thresholds=prior.thresholds,
                                                      type="truncatedMean", trim=0.05)

                        print(cellchat)

                        df.net = subsetCommunication(cellchat)

                        df.net
                    }
                ''')

        out = self.robjects.r['f'](adata, group_by)

        return out



import pandas as pd
import numpy as np
from itertools import combinations, chain
from collections import Counter
import time
from multiprocessing import Pool

from scipy.sparse import csr_matrix, coo_matrix

import scanpy as sc
from tqdm import tqdm
import networkx as nx

import seaborn as sns
import colorcet as cc
import matplotlib.path
from matplotlib import colors

import warnings

from .hotspot import get_all_hotspot_keys, compute_hotspot_boundary


def coexpression_hotspots(adata, jaccard_matrix=None, hotspot_lookup=None, threshold=0.7,
                          min_size=0, min_genes=5, cutoff=0.5, verbose=False,
                          processes=1, divisions=10, use_core=True, core_k=3, resolution=1.0):
    """
    Compute coexpression hotspots in a dataset for which some set of single-gene or interaction hotspots have already
    been computed.
    """
    if jaccard_matrix is None:
        if verbose:
            print("Computing hotspot similarity matrix.")
        jaccard_matrix, hotspot_lookup = hotspot_jaccard(adata, processes=processes,
                                                         verbose=verbose, divisions=divisions)

    try:
        multigene_hotspots = hotspot_groups(jaccard_matrix, hotspot_lookup, use_core=use_core, core_k=core_k,
                                            resolution=resolution,
                                            threshold=threshold, min_genes=min_genes)
    except RuntimeError as e:
        warnings.warn("No coexpression above threshold was detected. Try again with a lower threshold.")
        return

    median_sizes = []
    for idx, multigene_hotspot in enumerate(multigene_hotspots):
        sizes = []
        for key, v in multigene_hotspot:
            hotspot_size = np.count_nonzero(adata.obs[key] == v)
            sizes.append(hotspot_size)
        median_sizes.append(int(np.median(sizes)))
    multi_arrays = {}
    new_multigene_hotspots = []
    score_dict = {}
    for idx, (multigene_hotspot, median_size) in enumerate(
            sorted(zip(multigene_hotspots, median_sizes), key=lambda x: -x[1])):
        if median_size < min_size:
            break
        hotspots_per_spot = np.zeros(adata.shape[0])

        for idx2, (key, v) in enumerate(multigene_hotspot):
            hotspots_per_spot += (adata.obs[key] == v).astype(np.int_)

        if cutoff is None:
            inds = np.argsort(hotspots_per_spot)[::-1][:median_size]
        else:
            inds = np.where(hotspots_per_spot > cutoff * len(multigene_hotspot))[0]

        score = np.sum(hotspots_per_spot[inds]) / np.sum(hotspots_per_spot)
        score_dict[idx] = score
        # compute essence of overlap

        v = np.full(adata.shape[0], np.nan)
        v[inds] = 0
        multi_arrays[idx] = pd.Categorical(v)
        new_multigene_hotspots.append(multigene_hotspot)

    multi_arrays = {f"hotspots_multi_{idx}": v for idx, (k, v) in enumerate(multi_arrays.items())}
    # adata.obs = pd.concat([adata.obs, pd.DataFrame(multi_arrays, index=adata.obs.index)], axis=1)
    adata.obs.drop(adata.obs.filter(regex='hotspots_multi').columns.tolist(), axis=1, inplace=True)
    tmp_df = adata.obs.merge(pd.DataFrame(multi_arrays, index=adata.obs.index), how="outer",
                             left_index=True,
                             right_index=True, suffixes=('_old', None))
    tmp_df.drop(tmp_df.filter(regex='_old$').columns.tolist(), axis=1, inplace=True)
    adata.obs = tmp_df
    new_multigene_hotspots = {str(idx): v for idx, v in enumerate(new_multigene_hotspots)}
    adata.uns["multi_hotspots"] = new_multigene_hotspots

    # Save standard colors for each coexpression hotspot
    cm1 = sns.color_palette(cc.glasbey_bw, n_colors=len(adata.uns['multi_hotspots']))
    adata.uns['multi_hotspots_colors'] = {k: colors.to_hex(cm1[int(k)]) for k in adata.uns['multi_hotspots'].keys()}

    return


def multi_hotspot_groupings(adata):
    multi_hotspots = [v for v in adata.obs if "hotspots_multi" in v]
    multi_hotspots_df = sc.get.obs_df(adata, keys=multi_hotspots)
    # compute to matrix, 0 -> 1, nan -> 0
    multi_hotspots_array = np.nan_to_num(np.array(multi_hotspots_df, dtype=np.float32) + 1).astype(
        np.int_)
    type_counter = Counter()
    for idx in range(adata.shape[0]):
        t = np.where(multi_hotspots_array[idx, :])
        type_counter[t] += 1

    return sorted(type_counter.items(), key=lambda x: -x[1])


def hotspot_decomposition(adata, compute_type_counts=False, neg_weight=1.0):
    # assumes that single-gene hotspots and multi-gene hotspots have already been computed
    all_hotspots = [v for v in adata.obs if "hotspots_" in v and "multi" not in v]
    multi_hotspots = [v for v in adata.obs if "hotspots_multi" in v]

    multi_hotspots_df = sc.get.obs_df(adata, keys=multi_hotspots)
    # compute to matrix, 0 -> 1, nan -> 0
    multi_hotspots_array = np.nan_to_num(np.array(multi_hotspots_df, dtype=np.float32) + 1).astype(
        np.int_)

    weights_dict = hotspot_decomposition_keys(adata, all_hotspots, multi_hotspots_array,
                                              neg_weight=neg_weight)

    type_counter = Counter()
    if compute_type_counts:
        types = [tuple(np.where(weights)[0]) for (weights, _) in weights_dict.values() if
                 np.sum(weights) > 0]
        for t in types:
            for subset in powerset(t):
                type_counter[subset] += 1
        return weights_dict, sorted(type_counter.items(), key=lambda x: -x[1])

    return weights_dict


def hotspot_decomposition_keys(adata, keys, multi_hotspots_array, neg_weight=None):
    weights_dict = {}
    for hotspot in tqdm(keys):
        hotspot_vector = np.array(adata.obs[hotspot], dtype=np.float32)
        hotspot_vector[pd.notnull(hotspot_vector)] = 1
        hotspot_vector = np.nan_to_num(np.array(hotspot_vector, dtype=np.float32)).astype(np.int_)

        weights, score = hotspot_decomposition_sub(hotspot_vector, multi_hotspots_array,
                                                   neg_weight=neg_weight)
        hotspot_name = hotspot[9:]
        weights_dict[hotspot_name] = (weights, score)

    return weights_dict


def hotspot_decomposition_sub(hotspot_vector, multi_hotspots_array, neg_weight=None):
    if neg_weight is None:
        neg_weight = 1.0
    original_size = np.sum(hotspot_vector.flatten()).astype(np.float64)
    weights = np.zeros(multi_hotspots_array.shape[1])
    score = 0.0

    while True:
        overlap = np.sum(np.multiply(hotspot_vector.reshape(-1, 1), multi_hotspots_array), axis=0)
        neg_overlap = np.sum(np.multiply(1 - hotspot_vector.reshape(-1, 1), multi_hotspots_array),
                             axis=0)
        score_change = overlap - neg_weight * neg_overlap
        max_ind = np.argmax(score_change)
        if score_change[max_ind] > 0:
            weights[max_ind] = 1
            intersects = np.logical_and(hotspot_vector, multi_hotspots_array[:, max_ind])
            hotspot_vector[intersects] = 0
            score += score_change[max_ind]
        else:
            break

    score /= original_size
    # check if we met the minimum coverage
    # if np.sum(hotspot_vector.flatten())/original_size < min_covered_frac:
    # we didn't so return nothing
    #    return np.zeros(multi_hotspots_array.shape[1]), 0.0
    return weights, score


def hotspot_jaccard(adata, keys=None, processes=4, verbose=False, divisions=5):
    if keys is None:
        keys = get_all_hotspot_keys(adata)

    # Construct a sparse binary assignment matrix representing which spots are members of which hotspots in
    # CSR format
    num_hotspots = 0
    mean_size = 0
    indptr_counter = 0
    offsets = {}

    indices_list = []
    indptr = [0]

    hotspot_sizes = []

    hotspot_lookup = []

    for idx, hotspot_key in enumerate(keys):
        offsets[hotspot_key] = num_hotspots
        v = pd.Categorical(adata.obs[hotspot_key])
        n_hotspots = len(v.categories)
        for c in range(n_hotspots):
            inds = np.where(v.codes == c)[0]
            indices_list.append(inds)
            indptr_counter += len(inds)
            indptr.append(indptr_counter)
            hotspot_sizes.append(len(inds))
            mean_size += len(inds)
            hotspot_lookup.append((hotspot_key, v.categories[c]))

        num_hotspots += n_hotspots

    indices = np.concatenate(indices_list)
    data = np.ones(indices.shape)
    hotspot_matrix = csr_matrix((data, indices, indptr)).tocsc()

    hotspot_sizes = np.array(hotspot_sizes, dtype=np.float32)

    # Iterate over columns, each representing a spot and the hotspots it belongs to
    num_pairs = 0
    combination_iterators = []
    num_pairs_arr = []
    for col in range(hotspot_matrix.shape[1]):
        hotspots = hotspot_matrix[:, col].nonzero()[0]
        l = len(hotspots)
        num_pairs += l * (l - 1) // 2
        num_pairs_arr.append(l * (l - 1) // 2)
        # Append to the running list an iterator that iterates over every pair of hotspots that share the current
        # element
        combination_iterators.append(combinations(hotspots, 2))

    # TODO: maybe handle the case we can't import multiprocessing (does that happen?)
    if processes is not None:
        p = Pool(processes)
    # Spatially divide the spots
    coords = adata.obsm['spatial']
    xmin, ymin = np.min(coords, axis=0)
    xmax, ymax = np.max(coords, axis=0)
    xmax += 0.00001
    ymax += 0.00001
    xdiff = xmax - xmin
    ydiff = ymax - ymin
    # Partition up all the spots based on a grid
    data_list = []
    for idx in range(divisions):
        for idy in range(divisions):
            cur_xmin = xmin + (idx / divisions) * xdiff
            cur_xmax = xmin + ((idx + 1) / divisions) * xdiff

            cur_ymin = ymin + (idy / divisions) * ydiff
            cur_ymax = ymin + ((idy + 1) / divisions) * ydiff
            in_current_element = np.all([coords[:, 0] >= cur_xmin, coords[:, 0] < cur_xmax,
                                         coords[:, 1] >= cur_ymin, coords[:, 1] < cur_ymax],
                                        axis=0)
            inds = np.where(in_current_element)[0]
            if len(inds) == 0:
                continue
            cur_iters = []
            cur_num_pairs = 0
            for ind in inds:
                cur_iters.append(combination_iterators[ind])
                cur_num_pairs += num_pairs_arr[ind]
            data_list.append((cur_iters, cur_num_pairs, num_hotspots))

    num_tasks = len(data_list)

    overlap_matrix = csr_matrix((num_hotspots, num_hotspots))

    if verbose:
        list_wrapper = tqdm
    else:
        list_wrapper = lambda x, total: x

    if processes is not None:
        for sub_matrix in list_wrapper(p.imap_unordered(_multi_wrapper, data_list),
                                       total=num_tasks):
            overlap_matrix += sub_matrix
            del sub_matrix
    else:
        for sub_matrix in list_wrapper((_multi_wrapper(data) for data in data_list),
                                       total=num_tasks):
            overlap_matrix += sub_matrix
            del sub_matrix

    if processes is not None:
        p.close()

    del hotspot_matrix
    overlap_matrix = overlap_matrix.tocoo()

    # Compute the jaccard matrix
    if verbose:
        print("Computing jaccard matrix")
    row_hotspot_size = hotspot_sizes[overlap_matrix.row]
    col_hotspot_size = hotspot_sizes[overlap_matrix.col]
    overlap_count = overlap_matrix.data
    jaccard_matrix_data = row_hotspot_size + col_hotspot_size
    jaccard_matrix_data -= overlap_count
    jaccard_matrix_data = overlap_count / jaccard_matrix_data
    if verbose:
        print("Converting jaccard matrix")
    jaccard_matrix = coo_matrix(
        (jaccard_matrix_data, (overlap_matrix.row, overlap_matrix.col)),
        overlap_matrix.shape).tocsr()

    return jaccard_matrix, hotspot_lookup


def _multi_wrapper(data):
    iters, num_pairs, num_hotspots = data
    return compute_overlap_matrix(iters, num_pairs, num_hotspots, verbose=False)


def compute_overlap_matrix(iterator_list, num_pairs, num_hotspots, verbose=True):
    t0 = time.time()
    idx = 0
    if verbose:
        print(f"Constructing overlap matrix with {num_pairs} pairs.")
    row = np.empty(num_pairs, dtype=np.int32)
    col = np.empty(num_pairs, dtype=np.int32)
    if verbose:
        iterator_list = tqdm(iterator_list)
    for it in iterator_list:
        for i, j in it:
            row[idx], col[idx] = i, j
            idx += 1
    data = np.ones(num_pairs, dtype=np.float32)
    if verbose:
        print("Converting overlap matrix")
    overlap_matrix = coo_matrix((data, (row, col)), (num_hotspots, num_hotspots),
                                dtype=np.float32).tocsr()
    del data, row, col
    if verbose:
        t1 = time.time()
        print(f"Completed in {t1 - t0} seconds.")
    return overlap_matrix


def hotspot_groups(jaccard_matrix, hotspot_lookup, threshold, min_genes=5, use_core=True, core_k=3,
                   resolution=1.0):
    jaccard_matrix_coo = jaccard_matrix.tocoo()
    above_threshold = jaccard_matrix_coo.data > threshold

    if np.count_nonzero(above_threshold) == 0:
        raise RuntimeError("No Jaccard similarity values detected above threshold.")
    row = jaccard_matrix_coo.row[above_threshold]
    col = jaccard_matrix_coo.col[above_threshold]
    data = jaccard_matrix_coo.data[above_threshold]
    jaccard_matrix_coo.data = data
    jaccard_matrix_coo.row = row
    jaccard_matrix_coo.col = col
    jaccard_matrix_reduced = jaccard_matrix_coo.tocsr()
    del jaccard_matrix_coo


    g = nx.from_scipy_sparse_matrix(jaccard_matrix_reduced)
    if use_core:
        g = nx.k_core(g, k=core_k)

    if g.number_of_edges() == 0:
        raise RuntimeError("Not enough Jaccard similarity values detected above threshold.")

    multigene_hotspots = []
    vl = nx.algorithms.community.louvain_communities(g, resolution=resolution)
    for c in sorted(vl, key=lambda x: -len(x)):
        if len(c) < min_genes:
            continue
        multigene_hotspot = [hotspot_lookup[i] for i in c]
        multigene_hotspots.append(multigene_hotspot)

    return multigene_hotspots


def compute_multi_boundaries(adata, alpha_max, alpha_min, ids=None):
    multi_hotspots = [v for v in adata.obs if "hotspots_multi" in v]

    adata.uns["multi_boundaries"] = {}
    for idx, hotspot_key in tqdm(enumerate(multi_hotspots), total=len(multi_hotspots)):
        if ids is not None and idx not in ids:
            continue
        boundary = compute_hotspot_boundary(adata, hotspot_key, 0, alpha_max=alpha_max,
                                            alpha_min=alpha_min)
        # saving of anndata converts dictionary indices in .uns to strings anyway
        adata.uns["multi_boundaries"][str(idx)] = boundary


def multi_closure(adata):
    hotspot_keys = [v for v in adata.obs if "hotspots_multi" in v]
    points = adata.obsm['spatial']
    for idx, hotspot_key in enumerate(hotspot_keys):
        boundary = adata.uns["multi_boundaries"][idx]
        path = matplotlib.path.Path(boundary)
        v = pd.Categorical(path.contains_points(points, radius=-180), categories=(True,))
        v.categories = [0]
        adata.obs[hotspot_key] = v


def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(1, len(s) + 1))



import numpy as np
import pandas as pd
import scanpy as sc
from tqdm import tqdm
import leidenalg as la
import igraph as ig
try:
    from scipy.sparse import coo_array
except ImportError:
    # google colab compatibility
    from scipy.sparse import coo_matrix as coo_array
from scipy.spatial import KDTree


def segmentation_3d(adata, z_key="Bregma", output_key="segmentation_3d", inplace=True):
    """
    :param adata: Anndata object
    :param z_key: Column name in adata.obs containing the z-value associated with each observation.
    :param output_key: Column name in adata.obs to store segmentation membership in
    :param inplace: If `True`, modify adata. If `False` return a copy.

    :return: adata object with segmentation added to adata.obs, only if `inplace=False`.
    """
    if not inplace:
        adata = adata.copy()
    kv = 20
    k_inter = 5
    gamma = 1
    omega = 0.0001
    alpha = 0.2

    layer_offset = 0
    layer_offset_list = [0]

    within_adj_list = []
    inter_layer_edges = []

    # PCA reduction over all layers
    n_pca = 8
    sc.tl.pca(adata, n_comps=n_pca)

    z_values = np.unique(adata.obs[z_key])

    for layer in tqdm(range(12)):
        adata_sub_1 = adata[adata.obs['Bregma'] == z_values[layer], :]
        X_pca = adata_sub_1.obsm['X_pca']
        xs = adata_sub_1.obsm['spatial']
        kd_1 = KDTree(xs)
        dist_1, nn_1 = kd_1.query(x=xs, k=range(2, kv + 2))
        dist_weights = np.exp(-(dist_1 / 0.04))

        kd_pca = KDTree(X_pca)
        dist_pca, nn_pca = kd_pca.query(x=X_pca, k=range(2, kv + 2))
        exp_weights = np.exp(-np.linalg.norm(X_pca.reshape((-1, 1, n_pca)) - X_pca[nn_1, :], axis=-1) / 2)

        # compute within_layer adjacency
        i = np.ravel([(i,) * kv for i in range(len(xs))])
        j = np.ravel(nn_1)
        data = np.ravel(dist_weights)
        within_adj_dist = coo_array((data, (i + layer_offset, j + layer_offset))).tocsr()

        i = np.ravel([(i,) * kv for i in range(len(xs))])
        j = np.ravel(nn_pca)
        data = np.ravel(exp_weights)
        within_adj_expr = coo_array((data, (i + layer_offset, j + layer_offset))).tocsr()

        within_adj = (alpha * within_adj_dist + (1 - alpha) * within_adj_expr).tocoo()

        within_adj_list.append(within_adj)

        # add to between layer adjacency
        if layer < 11:
            adata_sub_2 = adata[adata.obs['Bregma'] == z_values[layer + 1], :]
            xt = adata_sub_2.obsm['spatial']
            kd_2 = KDTree(xt)

            dist_12, nn_12 = kd_1.query(x=xt, k=k_inter)
            dist_21, nn_21 = kd_2.query(x=xs, k=k_inter)

            # C = cdist(adata_sub_1.obsm['spatial'], adata_sub_2.obsm['spatial'], metric='sqeuclidean')
            A = np.zeros((len(xs), len(xt)))

            for i, row in enumerate(nn_21):
                for j in row:
                    weight = 1 / k_inter
                    inter_layer_edges.append((i + layer_offset, j + layer_offset + len(xs), omega * weight))
                    A[i, j] = 1
            for j, row in enumerate(nn_12):
                for i in row:
                    weight = 1 / k_inter
                    inter_layer_edges.append((i + layer_offset, j + layer_offset + len(xs), omega * weight))
                    A[i, j] = 1

        layer_offset += len(xs)
        layer_offset_list.append(layer_offset)

    partitions = []
    for adj in within_adj_list:
        adj.resize((layer_offset, layer_offset))
        g = ig.Graph.Weighted_Adjacency(adj)
        g.vs['node_size'] = 1
        partitions.append(la.RBConfigurationVertexPartition(g, weights='weight', resolution_parameter=gamma))
        # partitions.append(la.CPMVertexPartition(g, weights='weight', resolution_parameter=gamma,
        #                                       node_sizes='node_size'))

    inter_layer_edges_arr = np.array(inter_layer_edges).T
    data = inter_layer_edges_arr[2, :]
    i = inter_layer_edges_arr[0, :].astype(np.int_)
    j = inter_layer_edges_arr[1, :].astype(np.int_)
    inter_layer_adj = coo_array((data, (i, j)))
    inter_layer_adj.resize(layer_offset, layer_offset)
    inter_g = ig.Graph.Weighted_Adjacency(inter_layer_adj)
    inter_g.vs['node_size'] = 0
    partitions.append(la.CPMVertexPartition(inter_g, weights='weight', resolution_parameter=0.0,
                                            node_sizes='node_size'))

    optimiser = la.Optimiser()

    max_iters = 3
    for _ in tqdm(range(max_iters)):
        diff = optimiser.optimise_partition_multiplex(partitions)
        if diff == 0:
            break

    full_membership = np.array(partitions[0].membership)

    adata.obs[output_key] = pd.Categorical(full_membership)

    if not inplace:
        return adata



import scanpy as sc
import pandas as pd
import numpy as np
import anndata
import warnings
from tqdm import tqdm

from scipy.sparse import csr_matrix
from sklearn.neighbors import BallTree


def get_activity_dataframe(adata):
    obs_names = list(adata.obs)
    activity_names = [v for v in obs_names if "activity_" in v]
    return sc.get.obs_df(adata, keys=activity_names)


def write(adata, name):
    adata = adata.copy()
    adata.obs.to_pickle(f'{name}.pkl')
    adata.obs = pd.DataFrame(index=adata.obs.index)
    adata.write(f'{name}.h5ad')


def read(name):
    adata = anndata.read_h5ad(f'{name}.h5ad')
    adata.obs = pd.read_pickle(f'{name}.pkl')
    return adata


def spatial_smoothing(adata, mode=None):
    smoothing_matrix = get_neighbor_adjacency(adata.obsm['spatial'], eps=30)
    data = adata.X
    data2 = np.empty(data.shape)
    for idx in tqdm(range(smoothing_matrix.shape[0])):
        row = smoothing_matrix[idx, :]
        subdata = data[row.nonzero()[1], :].toarray()
        q1 = np.quantile(subdata, 0.2, axis=0)
        q2 = np.quantile(subdata, 0.8, axis=0)
        # print(q1.shape, q2.shape)
        if mode == "visium":
            data2[idx, :] = np.mean(subdata, axis=0)
        else:
            data2[idx, :] = np.mean([q1, q2], axis=0)
    adata.X = csr_matrix(data2)


def get_neighbor_adjacency(coords, eps=None, k=None, weight=None, return_row_col=False):
    """
    Compute either k-nearest neighbor or epsilon-radius neighbor adjacency graph
    :param coords: Spatial coordinates of each cell
    :param eps: radius cutoff for epsilon-radius neighbors
    :param k: number of nearest neighbors for k-nearest neighbors
    :param weight: Function mapping distance to value in weighted adjacency (default: f(x) = 1)
    :param return_row_col: If true, return the row and column variables from sparse coo format
    :return:
    """
    tree = BallTree(coords)
    n_cells = len(coords)
    if eps is not None and k is not None:
        warnings.warn('k is ignored if eps is provided')

    if eps is not None:
        nearby_col, dist = tree.query_radius(coords, eps, return_distance=True)
        coords = []
        nearby_row = []
        for i, col in enumerate(nearby_col):
            nearby_col[i] = col
            if weight is not None:
                coords.append(weight(dist[i]))
            nearby_row.append(i * np.ones(shape=col.shape, dtype=col.dtype))
    elif k is not None:
        nearby_col = tree.query(coords, k,
                                return_distance=False)  # [:, 1:]  # remove self connections
        nearby_row = []
        for i, col in enumerate(nearby_col):
            nearby_row.append(i * np.ones(shape=col.shape, dtype=col.dtype))
    else:
        raise ValueError("Provide either distance eps or number of neighbors k")

    nearby_row = np.concatenate(nearby_row)
    nearby_col = np.concatenate(nearby_col)
    if weight is None or k is not None:
        coords = np.ones(shape=nearby_row.shape)
    else:
        coords = np.concatenate(coords)

    nearby = csr_matrix((coords, (nearby_row, nearby_col)),
                        shape=(n_cells, n_cells))

    if return_row_col:
        return nearby, nearby_row, nearby_col

    return nearby



import numpy as np
from collections import Counter

from scipy.cluster.hierarchy import linkage
from scipy.spatial.distance import squareform

from nest.hotspot import differential_expression, hotspot_decomposition

import pandas as pd


def geometry_scores(adata):
    # Compute a significance score based on each hotspot and how much stuff is consistent
    ncells = adata.shape[0]
    arr = np.zeros(ncells)
    arr_unique = np.zeros(ncells)
    arr_unique_expr = np.zeros(ncells)
    X = adata.X.toarray()
    var_names = np.array([v.casefold() for v in adata.var_names])
    for idx, v in adata.uns['multi_hotspots'].items():
        # subarr = np.zeros(ncells)
        for key, c in v:
            gene_idx = np.where(var_names == key[9:].casefold())[0]
            sub = (adata.obs[key] == int(c)).to_numpy().astype(np.int_).reshape(-1)
            arr += sub
            arr_unique += sub.astype(np.float_) / adata.obs[key].count()
            expr = X[:, gene_idx].reshape(-1)
            if expr.size == 0:
                continue
            arr_unique_expr[sub.astype(np.bool_)] += expr[sub.astype(np.bool_)] / np.sum(expr)
    # normalize
    arr = np.log(1 + arr)
    arr /= np.max(arr)
    arr_unique /= np.max(arr_unique)
    arr_unique_expr /= np.max(arr_unique_expr)
    score_dict = {"geometry": arr,
                  "geometry_unique": arr_unique,
                  "geometry_unique_expr": arr_unique_expr}
    geometry_score_df = pd.DataFrame(score_dict, index=adata.obs.index)
    adata.obs = pd.concat([adata.obs, geometry_score_df], axis=1)
    return arr, arr_unique, arr_unique_expr


def geometric_markers(adata, inds, min_fc=1, use_decomp=False):
    # find all genes in exactly one of the hotspots
    gene_counts = Counter()

    if use_decomp:
        weights_dict = hotspot_decomposition(adata)

    for ind in inds:
        if use_decomp:
            genes = [k for k, v in weights_dict.items() if v[0][ind] and '_' not in k]
        else:
            genes = [k[9:] for k, v in adata.uns['multi_hotspots'][str(ind)]]
        #print(genes)
        for gene in genes:
            if "_" in gene:
                # Need to skip over any interaction hotspots involved in here
                continue
            gene_counts[gene] += 1

    # unique genes are in exactly one multi hotspot
    #print(gene_counts)
    unique_genes = {k for k, v in gene_counts.items() if v == 1}

    markers = {k: [] for k in inds}
    for ind in inds:
        if use_decomp:
            genes = [k for k, v in weights_dict.items() if v[0][ind] and '_' not in k]
        else:
            genes = [k[9:] for k, v in adata.uns['multi_hotspots'][str(ind)]]
        for gene in genes:
            if gene in unique_genes:
                markers[ind].append(gene)
    #print(markers)
    # Now do DE filtering to find the ones that are actually more highly expressed
    for cur_ind in inds:
        inds_a = np.where(pd.notnull(adata.obs[f'hotspots_multi_{cur_ind}']))[0]
        adata_sub = adata.copy()
        adata_sub.obs = pd.DataFrame(index=adata_sub.obs.index)
        adata_sub = adata_sub[:, markers[cur_ind]]
        gene_set = set(markers[cur_ind])
        gene_fc = np.full(fill_value=0, shape=len(gene_set), dtype=np.float_)
        for other_ind in inds:
            if cur_ind == other_ind:
                continue
            inds_b = np.where(pd.notnull(adata.obs[f'hotspots_multi_{other_ind}']))[0]
            de = differential_expression(adata_sub, inds_a, inds_b, use_raw=False)
            # positively_expressed = de.index[de['log2(fc)'] > min_fc]
            # gene_set = gene_set & set(positively_expressed)
            var = np.array(de['log2(fc)'])
            try:
                gene_fc += var
            except ValueError:
                print(gene_fc.shape, gene_fc)
                print(var.shape, var)
        genes_by_fc = [v[0] for v in sorted(zip(markers[cur_ind], gene_fc), key=lambda x: -x[1])]
        markers[cur_ind] = genes_by_fc

    return markers


def similarity_map(adata, idx, ax=None, linewidth=0.5, linecolor="black", title=""):
    ncells = adata.shape[0]
    arr = np.zeros(ncells)
    var_names = np.array([v.casefold() for v in adata.var_names])
    for key, c in adata.uns['multi_hotspots'][str(idx)]:
        sub = pd.notnull(adata.obs[key]).to_numpy().astype(np.int_).reshape(-1)
        arr += sub
    # normalize
    arr /= np.max(arr)

    return arr


def sim_linkage(adata, method='single'):
    # Compute pairwise similarity between all clusters
    n_ch = len(adata.uns['multi_hotspots'])
    pairwise_sim = np.zeros(shape=(n_ch, n_ch))
    for idx in adata.uns['multi_hotspots'].keys():
        sim = similarity_map(adata, idx)
        pairwise_sim[int(idx), int(idx)] = 1
        for idy in adata.uns['multi_hotspots'].keys():
            if idx == idy:
                continue
            val = np.mean(sim[pd.notnull(adata.obs[f'hotspots_multi_{idy}'])])
            pairwise_sim[int(idx), int(idy)] += val
            # pairwise_sim[int(idy), int(idx)] += val

    pairwise_sim = np.minimum(pairwise_sim, pairwise_sim.T)
    pairwise_dist = 1 - pairwise_sim
    Z = linkage(squareform(pairwise_dist), method=method)
    return pairwise_sim, Z



from collections import deque
import scipy.sparse
import scipy.spatial
import scipy.stats
from tqdm import tqdm

from nest import get_neighbor_adjacency

import numpy as np
import pandas as pd

from sklearn.decomposition import PCA
from sklearn.cluster import KMeans


def hmrf(adata, regions, eps, max_iterations=200, verbose=False, dim=16, pca=True, beta=1):
    ncells = len(adata)

    # perform PCA on genes
    X = np.array(adata.X.toarray())
    if pca:
        pca_transform = PCA(n_components=dim)
        pca_transform.fit(X)
        pca = pca_transform.transform(X)
    else:
        if adata.shape[1] != dim:
            raise ValueError
        pca = X

    pca_t = np.transpose(pca)





    gamma = initialize_gamma()

    last_10_ll = deque()

    flag = False  # convergence stopping

    if max_iterations is not None:
        iterations = max_iterations
    else:
        iterations = max_iterations

    if verbose:
        pbar = tqdm(range(iterations), leave=False)
    else:
        pbar = range(iterations)

    for it in pbar:
        ll = np.zeros(shape=(ncells, regions))

        # Compute log-likelihood summed across all models
        for i, model in enumerate(likelihood_models):
            model.update_parameters(gamma)

            if model_weights is not None:
                weight = model_weights[i]
            else:
                weight = 1

            ll += weight * model.compute_log_likelihood(gamma)

        # Also do class size-based likelihood
        class_prob = np.mean(gamma, axis=0)
        ll += np.log(class_prob).reshape(1, regions)

        # Compute new class probabilities
        ll = ll
        gamma = scipy.special.softmax(ll, axis=1)

        model_ll = np.mean(ll * gamma)

        if tol is not None and it >= 10 and model_ll - max(last_10_ll) < tol:
            flag = True

        if len(last_10_ll) >= 10:
            last_10_ll.popleft()
        last_10_ll.append(model_ll)

        # compute overall model likelihood to watch for convergence
        class_ll = ll[:, class_labels].mean()

        if verbose:
            pbar.set_description(f"LL: {model_ll}")

        if flag:
            converged = True
            if verbose:
                pbar.set_description("Converged: LL %f" % model_ll)
            break

    if update_labels:
        adata.obs[label_name] = pd.Categorical(class_labels)



def initialize_gamma(adata, regions):
    try:
        data = adata.X.toarray()
    except TypeError:
        data = adata.X

    # do dimensionality reduction with PCA
    reduced_dims = 16
    pca = PCA(n_components=reduced_dims)
    data = pca.fit_transform(data)

    class_labels = KMeans(n_clusters=regions).fit(data).labels_
    labels_onehot = pd.get_dummies(class_labels, columns=range(regions)).to_numpy()

    # how much uncertainty to introduce in initial clusters
    eps = 0.1

    gamma = (1 - eps) * labels_onehot + eps / regions * np.ones(
        shape=labels_onehot.shape)

    return gamma


def update_parameters(gamma, pca):
    regions = gamma.shape[1]
    dim = pca.shape[1]
    gamma_sum = np.sum(gamma, axis=0)

    gamma_t = np.transpose(gamma)
    means = (gamma_t @ pca) / gamma_sum.reshape(regions, 1)

    pca_minus_mean = pca.reshape((-1, 1, dim)) - means.reshape(
        (1, regions, dim))
    A = pca_minus_mean.reshape(-1, regions, dim, 1) * \
        pca_minus_mean.reshape(-1, regions, 1, dim)

    covs = np.einsum('ij,jimn->imn', gamma_t, A) / gamma_sum.reshape((regions, 1, 1))

    return means, covs


import warnings
import scipy.sparse

import numpy as np

from sklearn.preprocessing import normalize
from scipy.spatial import Delaunay



def compute_spatial_transport_matrices(adata,
                                       secreted_std,
                                       secreted_threshold_cutoff,
                                       contact_threshold=None,
                                       z_key=None):
    """
    Return the spatial prior belief of possibility of interaction between every possible pair of
    cells for both secreted signaling and cell-cell contact type interactions
    """

    ncells = adata.shape[0]

    # gaussian diffusion kernel - threshold is two standard deviations
    sigma = secreted_std

    def weight_fn(x):
        return np.exp(-0.5 * ((x / (2 * sigma)) ** 2))

    coords = adata.obsm['spatial']
    if z_key is not None:
        z = np.array(adata.obs[z_key]).reshape(-1, 1)
        coords = np.concatenate([coords, z], axis=1)
    transport_secreted = get_neighbor_adjacency(coords, eps=sigma * secreted_threshold_cutoff,
                                                weight=weight_fn)
    if z_key is None:
        # delauney for direct contact
        tri = Delaunay(adata.obsm['spatial'])
        indptr, indices = tri.vertex_neighbor_vertices
        data = np.ones(shape=indices.shape)
        transport_contact = scipy.sparse.csr_matrix((data, indices, indptr), shape=[ncells, ncells])

        # filter out edges in delauney triangulation that are too long
        if contact_threshold is not None:
            transport_contact = transport_contact.multiply(
                get_neighbor_adjacency(coords, eps=contact_threshold))

        # normalize across number of neighbors
        transport_contact = normalize(transport_contact, norm='l1', axis=1)
        adata.obsp['transport_contact'] = transport_contact

    # transport_secreted = normalize(transport_secreted, norm='l1', axis=1)


    adata.obsp['transport_secreted'] = transport_secreted



import os

import anndata
import pandas as pd
import scanpy as sc
import squidpy as sq
import numpy as np

from nest.utility import spatial_smoothing

SQUIDPY_DATASETS = ["seqfish", "slideseq", "merfish"]

VISIUM_DATASETS = [
    "V1_Breast_Cancer_Block_A_Section_1",
    "V1_Mouse_Brain_Sagittal_Anterior",
    "V1_Mouse_Brain_Sagittal_Posterior",
    "V1_Adult_Mouse_Brain_Coronal_Section_1", "V1_Breast_Cancer_Block_A_Section_2",
    "V1_Human_Heart", "V1_Human_Lymph_Node", "V1_Mouse_Brain_Sagittal_Posterior",
    "V1_Mouse_Brain_Sagittal_Posterior_Section_2", "V1_Mouse_Brain_Sagittal_Anterior",
    "V1_Mouse_Brain_Sagittal_Anterior_Section_2",
    "Targeted_Visium_Human_Cerebellum_Neuroscience", "Parent_Visium_Human_Cerebellum",
    "Targeted_Visium_Human_BreastCancer_Immunology",
    "Targeted_Visium_Human_OvarianCancer_Pan_Cancer",
    "Targeted_Visium_Human_OvarianCancer_Immunology"
    "V1_Human_Brain_Section_2",
    "V1_Adult_Mouse_Brain_Coronal_Section_1", "V1_Adult_Mouse_Brain_Coronal_Section_2",
    "Targeted_Visium_Human_SpinalCord_Neuroscience", "Parent_Visium_Human_SpinalCord",
    "Targeted_Visium_Human_Glioblastoma_Pan_Cancer", "Parent_Visium_Human_Glioblastoma",
    "Parent_Visium_Human_BreastCancer",
    "Parent_Visium_Human_OvarianCancer", "Targeted_Visium_Human_ColorectalCancer_GeneSignature",
    "Parent_Visium_Human_ColorectalCancer", "V1_Mouse_Kidney"]

SPATIAL_LIBD_DATASETS = ["Spatial_LIBD_%s" % item for item in
                         ["151507", "151671", "151673", "151509", "151510", "151669", "151670",
                          "151674", "151675",
                          "151672", "151676", "151508"]]


def get_squidpy_data(dataset, **kwargs):
    if dataset == "seqfish":
        adata = sq.datasets.seqfish()
        adata.uns['um_scale'] = 0.001
        adata.obs['class'] = adata.obs['celltype_mapped_refined']
    elif dataset == "merfish":
        adata = sq.datasets.merfish()
        adata.uns['um_scale'] = 0.001
        try:
            bregma = kwargs["bregma"]
        except KeyError:
            bregma = 1
        adata = adata[adata.obs.Bregma == bregma].copy()
        adata.obs['class'] = adata.obs['Cell_class']
        adata.raw = adata
        sc.pp.filter_cells(adata, min_counts=20)
    elif dataset == "slideseq":
        adata = sq.datasets.slideseqv2().raw.to_adata()
        adata.uns['um_scale'] = 1
        adata.obs["class"] = adata.obs["cluster"]

        spatial_smoothing(adata)

        # Rotate data so hippocampus is lengthwise "horizontal"
        X = adata.obsm['spatial']
        # Center on origin
        X -= np.mean(X, axis=0)
        theta = np.radians(20)
        R = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]]).T
        X = X @ R
        adata.obsm['spatial'] = X
    else:
        return ValueError("Dataset not recognized")

    return adata


def load_stereo_seq_data(dataset_dir):
    file = os.path.join(dataset_dir, "stereoseq", "adata.h5ad")
    adata = anndata.read_h5ad(file)
    return adata


def get_data(dataset, dataset_dir=None, normalize=True, **kwargs):
    if dataset in SQUIDPY_DATASETS:
        adata = get_squidpy_data(dataset=dataset, **kwargs)
    elif dataset == "stereoseq":
        adata = load_stereo_seq_data(dataset_dir)
    elif dataset in SPATIAL_LIBD_DATASETS:
        if dataset_dir is None:
            raise ValueError("For spatialLIBD data, must set dataset_dir to directory containing the data.")
        expr_dir = os.path.join(dataset_dir, dataset)
        h5ad_path = os.path.join(expr_dir, "adata.h5ad")
        try:
            adata = anndata.read_h5ad(h5ad_path)
        except FileNotFoundError:
            adata = sc.read_10x_mtx(expr_dir).copy()
            coord_fp = os.path.join(expr_dir, "spatial_coords.csv")
            coord_df = pd.read_csv(coord_fp).values.astype(float)

            adata.obsm['spatial'] = coord_df[:, [1, 0]]

            info_df = pd.read_csv(os.path.join(expr_dir, "spot_info.csv"))
            clusters = info_df["layer_guess_reordered"].values.astype(str)
            adata.obs['class'] = pd.Categorical(clusters)
            convert_ensembl_names(adata)

            # Cache this adata so we don't have to redo all this next time
            adata.write(h5ad_path)
    elif dataset in VISIUM_DATASETS:
        adata = sc.datasets.visium_sge(dataset, include_hires_tiff=False)
        adata.var_names_make_unique()
    else:
        raise ValueError("Unrecognized dataset")

    sc.pp.filter_genes(adata, min_counts=1)  # only consider genes with more than 1 count
    if normalize:
        sc.pp.normalize_total(adata)
    if 'log1p' not in adata.uns:
        sc.pp.log1p(adata)

    return adata


def convert_ensembl_names(adata):
    try:
        import mygene
    except ImportError:
        raise RuntimeError("Package mygene must be installed to run convert_ensembl_names")

    mg = mygene.MyGeneInfo()
    ens = adata.var_names
    res = mg.querymany(ens, scope="ensembl.gene", returnall=True)

    name_map = {}
    for item in res['out']:
        try:
            name_map[item['query']] = item['symbol']
        except KeyError:
            # No symbol was returned
            name_map[item['query']] = item['query']

    for gene in res['missing']:
        name_map[gene] = gene

    adata.var_names = pd.Index([name_map[gene] for gene in adata.var_names])


def default_parameters(dataset):
    if "V1_Breast_Cancer_Block_A" in dataset:
        neighbor_eps = 300
        min_samples = 4
        hotspot_min_size = 15
    elif dataset in VISIUM_DATASETS:
        neighbor_eps = 170
        min_samples = 4
        hotspot_min_size = 15
    elif dataset == "seqfish":
        neighbor_eps = 0.12
        min_samples = 10
        hotspot_min_size = 25
    elif dataset == "merfish":
        neighbor_eps = 0.06
        min_samples = 5
        hotspot_min_size = 5
    elif dataset == "slideseq":
        neighbor_eps = 75
        min_samples = 5
        hotspot_min_size = 50
    else:
        raise ValueError

    return neighbor_eps, min_samples, hotspot_min_size



from tqdm import tqdm
import numpy as np
import pandas as pd

from nest.hotspot.hotspot import _compute_cutoff, compute_hotspots
from nest.utility.utility import get_activity_dataframe
from nest.cci.cci import compute_activity


def interaction_hotspots(adata, verbose=False, interactions=None, save_activity=False,
                         method='permutation', log=False, K=None,
                         min_active_count=1, sig_threshold=0.95, z_key=None, **kwargs):
    # check if activity scores are available
    activity_matrix = get_activity_dataframe(adata)
    if len(list(activity_matrix)) == 0:
        # TODO: really should only compute activity for interactions in the filter if interactions
        # is not None
        try:
            um_scale = adata.uns['um_scale']
        except KeyError:
            # Try to get it automatically assuming Visium data with 55um spots
            try:
                dataset = list(adata.uns['spatial'].keys())[0]
                um_scale = adata.uns['spatial'][dataset]['scalefactors']['spot_diameter_fullres'] / 55
            except KeyError:
                raise ValueError("Please set adata.uns['um_scale'] to the length of one micrometer"
                                 " in the units of adata.obsm['spatial'].") from None

        secreted_std = 50 * um_scale
        contact_threshold = 20 * um_scale

        perform_permutations = method == 'permutation'
        activity_matrix = compute_activity(adata, secreted_std=secreted_std,
                                           contact_threshold=contact_threshold,
                                           sig_threshold=sig_threshold,
                                           perform_permutation=perform_permutations,
                                           save_activity=save_activity, verbose=verbose,
                                           min_active_count=min_active_count,
                                           interactions=interactions,
                                           z_key=z_key, K=K)

    cols = enumerate(list(activity_matrix))

    region_dict = {}

    for _, interaction in cols:
        if interactions is not None and interaction not in interactions:
            # interaction filter
            continue
        # modify this to account for possibly having to go over multiple z layers
        data = activity_matrix[interaction]
        if method == 'permutation':
            cutoff = adata.uns['activity_significance_cutoff'][interaction]
        else:
            cutoff = _compute_cutoff(data, log=log)
        inds = data > cutoff

        region_offset = 0
        if z_key is None:
            regions = compute_hotspots(adata=adata, input_data=np.where(inds)[0], return_regions=True, **kwargs)
        else:
            regions = -1 * np.ones(adata.shape[0])
            for val in np.unique(adata.obs[z_key]):
                cur_slice_inds = np.logical_and(inds, adata.obs[z_key] == val)
                regions_sub = compute_hotspots(adata=adata, input_data=np.where(cur_slice_inds)[0],
                                               return_regions=True, **kwargs)
                # combine together into one array
                out_inds = np.where(pd.notnull(regions_sub))[0]
                print(out_inds)
                regions[out_inds] = regions_sub[out_inds] + region_offset
                print(regions[regions > 0])
                region_offset += np.max(regions_sub[out_inds])
            regions = pd.Categorical(regions, categories=np.arange(1, np.max(regions) + 1))

        if regions is not None:
            region_dict[f"hotspots_{interaction}"] = regions

    hotspots_df = pd.DataFrame(region_dict, index=adata.obs.index)
    adata.obs = pd.concat([adata.obs, hotspots_df], axis=1)


if __name__ == "__main__":
    import squidpy as sq

    adata = sq.datasets.merfish()
    adata.uns['um_scale'] = 0.001
    adata.obs['z'] = adata.obs.Bregma * 0.01
    bregma_values = pd.unique(adata.obs.Bregma)

    neighbor_eps = 0.06
    min_samples = 5
    hotspot_min_size = 5

    adata = adata[:, ["Cck", "Cckar"]].copy()

    n_regions = interaction_hotspots(adata, eps=neighbor_eps, min_size=hotspot_min_size, min_samples=min_samples,
                         core_only=False, method="permutation", z_key='z', verbose=True)
    print(n_regions)



"""
Code for running other spatial segmentation methods
"""
import contextlib

import numpy as np
import pandas as pd
import sys


# TODO: do all the citations and stuff properly

class CellChat:
    def __init__(self):
        pass

    def run(self, adata, group_by="class"):
        # set up to run the R code
        import rpy2.robjects as robjects
        from rpy2.robjects.packages import importr
        import anndata2ri
        anndata2ri.activate()
        #from rpy2.robjects import pandas2ri
        #pandas2ri.activate()
        self.robjects = robjects
        self.anndata2ri = anndata2ri

        self.cellchat = importr("CellChat")

        # converter does not currently seem to support sparse matrix entries
        adata = adata.copy()
        try:
            adata.X = adata.X.todense()
        except AttributeError:
            pass

        # get rid of extraneous columns that may not convert properly anyway
        adata.obs = adata.obs.loc[:, [group_by]]

        self.robjects.r('''
                    f <- function(s, pos=NULL) {
                        assay(s, "logcounts") = assay(s, "X")

                        cellchat <- createCellChat(object = s, group.by = "%s")

                        print(cellchat)

                        CellChatDB <- CellChatDB.mouse

                        CellChatDB.use <- CellChatDB
                        cellchat@DB = CellChatDB.use

                        cellchat <- subsetData(cellchat)
                        cellchat <- identifyOverExpressedGenes(cellchat)
                        cellchat <- identifyOverExpressedInteractions(cellchat)
                        cellchat <- projectData(cellchat, PPI.mouse)

                        library(Matrix)
                        cellchat <- computeCommunProb(cellchat, type="truncatedMean", trim=0.05)

                        print(cellchat)

                        df.net = subsetCommunication(cellchat)

                        df.net
                    }
                ''' % group_by)

        out = self.robjects.r['f'](adata)
        df = pd.DataFrame(out).T
        df.columns = out.colnames
        return df

    def cellchat_score(self, adata, interaction, key="cellchat_score", group_by="class", rerun=False):
        try:
            if rerun:
                raise KeyError
            res = adata.uns['cellchat_res']
        except KeyError as e:
            res = self.run(adata, group_by)
            adata.uns['cellchat_res'] = res
        v = res[res['interaction_name_2'] == interaction][np.array(['target', 'prob'])].groupby('target').sum()
        # TODO: check that we have a nonzero number of rows in v
        v.index = adata.obs['class'].cat.categories[v.index - 1]
        # fill in any missing categories with 0 for next step lookup to work
        for s in set(np.unique(adata.obs[group_by])) - set(v.index):
            v.loc[s, 'prob'] = 0
        adata.obs[key] = np.array(v.loc[adata.obs[group_by], 'prob'])

class BayesSpace:
    """
    Bayesspace method from https://github.com/edward130603/BayesSpace
    """

    def __init__(self, regions, label_name="class_bayesspace"):
        self.regions = regions
        self.label_name = label_name

    def fit(self, adata, verbose=False, **kwargs):
        # set up to run the R code
        import rpy2.robjects as robjects
        from rpy2.robjects.packages import importr
        import anndata2ri
        anndata2ri.activate()
        self.robjects = robjects
        self.anndata2ri = anndata2ri

        self.sce = importr("SingleCellExperiment")
        self.bayesspace = importr("BayesSpace")

        platform = "Visium"

        # converter does not currently seem to support sparse matrix entries
        adatasub = adata.copy()
        adatasub.obs = pd.DataFrame({'row': adatasub.obs['array_row'],
                                    'col': adatasub.obs['array_col']},
                                    index=adatasub.obs.index)
        try:
            adatasub.X = adatasub.X.toarray()
        except AttributeError:
            pass

        adatasub.raw = None
        adatasub.obsp = None
        try:
            del adatasub.layers['exp']
        except KeyError:
            pass

        self.robjects.r('''
            f <- function(s, q) {
                assay(s, "logcounts") = assay(s, "X")
                s <- spatialPreprocess(s, platform="Visium", n.PCs=16, n.HVGs=2000, log.normalize=FALSE)
                s <- spatialCluster(s, q=q, platform="Visium", d=15, init.method="mclust", model="t", gamma=2, nrep=200, burn.in=100, save.chain=FALSE)
                print("Completed clustering")
                print(s)
                rowData(s)$is.HVG <- NULL
                metadata(s)$BayesSpace.data <- NULL
                s
            }
        ''')
        output = sys.stdout if verbose else None
        utils = importr('utils')
        out = self.robjects.r['f'](adatasub, self.regions)

        # copy the output labels back over
        adata.obs[self.label_name] = out.obs['spatial.cluster']

        return




class SpaGCN:
    def __init__(self, regions, label_name="class_spagcn"):
        self.regions = regions
        self.label_name = label_name

    def fit(self, adata, img=None, verbose=False, **kwargs):
        import SpaGCN as spg
        import scanpy as sc
        import torch
        import random
        import sys
        import contextlib

        output = sys.stdout if verbose else None
        adatasub = adata.copy()
        with contextlib.redirect_stdout(output):
            adatasub.var_names_make_unique()
            adatasub.X = np.exp(adatasub.X.toarray()) - 1
            spg.prefilter_genes(adatasub, min_cells=3)  # avoiding all genes are zeros
            spg.prefilter_specialgenes(adatasub)
            # Normalize and take log for UMI
            sc.pp.normalize_per_cell(adatasub)
            sc.pp.log1p(adatasub)

            # get adjacency matrix
            # Calculate adjacent matrix
            if img is not None:
                s = 1
                b = 49
                adj = spg.calculate_adj_matrix(x=x_pixel, y=y_pixel, x_pixel=x_pixel,
                                               y_pixel=y_pixel, image=img,
                                               beta=b,
                                               alpha=s, histology=True)
            else:
                x = adatasub.obsm['spatial'][:, 0]
                y = adatasub.obsm['spatial'][:, 1]
                adj = spg.calculate_adj_matrix(x=x, y=y, histology=False)

            p = 0.5
            # Find the l value given p
            l = spg.search_l(p, adj, start=0.001, end=1000, tol=0.01, max_run=100)

            # Set seed
            if False:
                r_seed = t_seed = n_seed = 100
                # Set seed
                random.seed(r_seed)
                torch.manual_seed(t_seed)
                np.random.seed(n_seed)
            # Search for suitable resolution
            # res = spg.search_res(adata, adj, l, n_clusters, start=0.7, step=0.1, tol=5e-3, lr=0.05, max_epochs=20,
            #                     r_seed=r_seed, t_seed=t_seed, n_seed=n_seed)

            clf = spg.SpaGCN()
            clf.set_l(l)

            # Run
            clf.train(adatasub, adj, init_spa=True, init="kmeans", n_clusters=self.regions, tol=5e-3,
                      lr=0.05,
                      max_epochs=200)
            y_pred, prob = clf.predict()
            adata.obs[self.label_name] = pd.Categorical(y_pred)

    @property
    def class_labels(self):
        return self.y_pred



class STLearn:
    def __init__(self):
        import stlearn
        self.stlearn = stlearn

    def fit(self, adata, img=None, **kwargs):
        st = self.stlearn

        data = st.convert_scanpy(adata)

        st.pp.filter_genes(data, min_cells=1)
        st.pp.normalize_total(data)
        st.pp.log1p(data)

        # run PCA for gene expression data
        st.em.run_pca(data, n_comps=50)

        data_SME = data.copy()
        # apply stSME to normalise log transformed data
        # with weights from morphological Similarly and physcial distance
        st.spatial.SME.SME_normalize(data_SME, use_data="raw",
                                     weights="weights_matrix_pd_md")
        data_SME.X = data_SME.obsm['raw_SME_normalized']
        st.pp.scale(data_SME)
        st.em.run_pca(data_SME, n_comps=50)

        # K-means clustering on stSME normalised PCA
        st.tl.clustering.kmeans(data_SME, n_clusters=17, use_data="X_pca", key_added="X_pca_kmeans")

        # louvain clustering on stSME normalised data
        st.pp.neighbors(data_SME, n_neighbors=20, use_rep='X_pca')
        st.tl.clustering.louvain(data_SME)

        self.stdata = data_SME
        print(self.stdata)

    @property
    def class_labels(self):
        raise NotImplementedError